// Copyright 2009 Christopher Smith
// 
// This file is part of smithers.extras.
// 
// smithers.extras is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
// 
// smithers.extras is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.extras. If not, see <http://www.gnu.org/licenses/>.

package smithers.extras;

/**
 * Provides a series of methods for converting between numeric values, and
 * {@code String}s. Supports the following:
 * <table>
 * <hr> <td>Type</td> <td>{@code long}</td> <td>{@code double}</td> </hr>
 * <tr> <td>Digits</td> <td>Supported</td> <td>Output only</td> </tr>
 * <tr> <td>Words</td> <td>Output only<br />Options: language, scale</td>
 * <td>Output only<br />Options: language, scale</td> </tr>
 * <tr> <td>Roman Numerals</td> <td>Output only</td> <td>Unsupported</td> </tr>
 * </table>
 * <table>
 * <hr> <td>Languages</td> </hr>
 * <tr> <td>English</td> </tr>
 * <tr> <td>French/Francais</td> </tr>
 * <tr> <td>German/Deutsch</td> </tr>
 * </table>
 * 
 * TODO: The following is out of date!
 * A {@code NumberString} object contains a number (either a {@code long} or a
 * {@code double}) as well as the options to be used. <br />
 * The following types of method are availiable:
 * <table>
 * <hr> <td>Type</td> <td>Prefix</td> <td>Function</td> </hr>
 * <tr> <td>Accessors</td> <td>{@code get}</td> <td>Get the number or
 * option.</td> </tr>
 * <tr> <td>Modifiers</td> <td>{@code set}</td> <td>Set the number or
 * option.</td> </tr>
 * <tr> <td>Encoders</td> <td>{@code to}</td> <td>Format the number in the
 * specified style.</td> </tr>
 * <tr> <td>Decoders</td> <td>{@code from}</td> <td>Parse a string in from the
 * specified style.</td> </tr>
 * </table>
 * 
 * The accessors and modifiers are only instance methods. The encoders are
 * instance methods, but static shortcut methods are availiable. The decoders
 * are currently only static methods.
 */
public class NumberString {
    
    /** English */
    public static final String ENGLISH = "en";
    /** French */
    public static final String FRENCH = "fr", FRANCAIS = "fr";
    /** German */
    public static final String GERMAN = "de", DEUTSCH = "de";
    
    /** The short scale. One billion = 10<sup>9</sup>. */
    public static final int SHORT_SCALE = 0;
    /**
     * The long scale. One billion = 10<sup>12</sup>. One thousand million =
     * 10<sup>9</sup>
     */
    public static final int LONG_SCALE_NO_MILLIARD = 1;
    /**
     * The long scale. One billion = 10<sup>12</sup>. One milliard =
     * 10<sup>9</sup>
     */
    public static final int LONG_SCALE_MILLIARD = 2;
    
    private String language;
    private boolean en = false;
    private boolean fr = false;
    private boolean de = false;
    
    private int scale;
    private boolean ss = false;
    private boolean lsnm = false;
    private boolean lsm = false;
    
    private char thousandSep;
    private char decimalSep;
    
    /**
     * Constructs a new {@code NumberString} object with the default properties.
     */
    public NumberString() {
        setOptions(ENGLISH);
    }
    
    /**
     * Constructs a new {@code NumberString} object with properties derived
     * from a specified language. The parameter should be one of the
     * language constants defined in this class.
     * 
     * @param language the language
     */
    public NumberString(String language) {
        setOptions(language);
    }
    
    /**
     * Constructs a new {@code NumberString} object with the specified
     * properties. The parameters should be chosen from the language and scale
     * constants defined in this class.
     * 
     * @param language the language
     * @param scale    the scale to use
     */
    public NumberString(String language, int scale) {
        setLanguage(language);
        setScale(scale);
    }
    
    /**
     * Gets the language of this {@code NumberString}.
     * 
     * @return the language
     */
    public String getLanguage() {
        return language;
    }
    
    /**
     * Gets the scale of this {@code NumberString}.
     * 
     * @return the scale
     */
    public int getScale() {
        return scale;
    }
    
    /**
     * Sets the all the options of this {@code NumberString} based on a
     * language. The parameter should be equal to one of the language constants
     * defined in this class.
     * 
     * @param language the language to derive the options from
     */
    public void setOptions(String language) {
        int scale=0;
        
        if (language.equals(ENGLISH)) scale = SHORT_SCALE;
        else if (language.equals(FRENCH)) scale = LONG_SCALE_MILLIARD;
        else if (language.equals(GERMAN)) scale = LONG_SCALE_MILLIARD;
        else throw new RuntimeException("Invalid or unsupported language");
        
        setLanguage(language);
        setScale(scale);
    }
    
    /**
     * Sets the language of this {@code NumberString}. The parameter should be
     * equal to one of the language constants defined in this class.
     * 
     * @param language the new language
     */
    public void setLanguage(String language) {
        if (language.equals(ENGLISH)) {
            en = true;
            fr = de = false;
            thousandSep = ',';
            decimalSep = '.';
        } else if (language.equals(FRENCH)) {
            fr = true;
            en = de = false;
            thousandSep = '.';
            decimalSep = ',';
        } else if (language.equals(GERMAN)) {
            de = true;
            en = fr = false;
            thousandSep = '.';
            decimalSep = ',';
        } else {
            throw new RuntimeException("Invalid or unsupported language");
        }
        
        this.language = language;
    }
    
    /**
     * Sets the scale of this {@code NumberString}. The parameter should be
     * equal to one of the scale constants defined in this class.
     * 
     * @param scale the new scale
     */
    public void setScale(int scale) {
        if (scale == SHORT_SCALE) {
            ss = true;
            lsnm = lsm = false;
        } else if (scale == LONG_SCALE_NO_MILLIARD) {
            lsnm = true;
            ss = lsm = false;
        } else if (scale == LONG_SCALE_MILLIARD) {
            lsm = true;
            ss = lsnm = false;
        } else {
            throw new RuntimeException("Invalid scale");
        }
        
        this.scale = scale;
    }
    
    /**
     * Converts a {@code long} value into a string of digits.
     * 
     * @param number the long value
     * @return the digits which represent {@code number} in base 10
     */
    public String toDigits(long number) {
        int com = 0;
        String result = "", minus = "";
        
        if (number == 0) return "0";
        if (number < 0) {
            minus = "-";
            number = -number;
        }
        while (number > 0) {
            if (com == 3) {
                result = thousandSep + result;
                com = 0;
            } else {
                com++;
            }
            result = (int)(number % 10) + result;
            number /= 10;
        }
        return minus + result;
    }
    
    /**
     * Converts a {@code double} value into a string of digits.
     * 
     * @param number the double value
     * @return the digits which represent {@code number} in base 10
     */
    public static String toDigits(double number) {
        // Cheat for now
        return new Double(number).toString();
//         String output = "", minus = "";
//         
//         if (number == 0) return "0";
//         if (number < 0) {
//             minus = "-";
//             number = -number;
//         }
//         
//         //Now what?
//         
//         return minus + output;
    }
    
    /**
     * Converts a string of digits into a {@code long} value.
     * 
     * @param text the string of digits
     * @return the {@code long} value of the string in base 10
     */
    public static long longFromDigits(String text) {
        long number = 0, minus = 1;
        char digit;
        long powerOfTen = 1;
        for (int i=0; i<text.length(); i++) {
            digit = text.charAt(i);
            switch (digit) {
                case '0': number *= 10; break;
                case '1': number *= 10; number += 1; break;
                case '2': number *= 10; number += 2; break;
                case '3': number *= 10; number += 3; break;
                case '4': number *= 10; number += 4; break;
                case '5': number *= 10; number += 5; break;
                case '6': number *= 10; number += 6; break;
                case '7': number *= 10; number += 7; break;
                case '8': number *= 10; number += 8; break;
                case '9': number *= 10; number += 9; break;
                case '-': minus *= -1;
            }
        }
        return number;
    }
    
    /**
     * Converts a {@code long} value into a string of words.
     * <ul>
     * <li>English: Supported</li>
     * <li>French: TODO: plurals</li>
     * <li>German: TODO: capitalisation, spaces</li>
     * </ul>
     * 
     * @param number the long value
     * @return a string of words representing {@code number}
     */
    public String toWords(long number) {
        String output = "", minus = "", temp = "", groupName = "";
        int curSec = 0, groupLength = 0, group = 0;
        boolean twoDigitFlag = false;
        
        if (number == 0) return (en ? "zero" : fr ? "zéro" : de ? "null" :"");
        if (number < 0) {
            minus = (en ? "minus " : fr ? "moins " : de ? "minus " :"");
            number = -number;
        }
        
        for (group = 0; number > 0; group += groupLength) {
            if (group==0) groupLength = 2;
            else if (group==2) groupLength = 1;
            else if (group==3) groupLength = 3;
            else if (lsnm) groupLength = 6;
            else groupLength = 3;
            
            groupName = getGroupName(group);
            
            curSec = (int)(number % (long)Math.pow(10, groupLength));
            number /= (long)Math.pow(10, groupLength);
            if (curSec > 0) {
                temp = getWordsPart(curSec) + groupName;
                if (output.equals("")) {
                    output = temp;
                } else if (twoDigitFlag) {
                    output = temp + (en ? " and " : fr ? " " : de ? "" :"") + output;
                } else {
                    output = temp + (en || fr ? ", " : de ? "" :"") + output;
                }
                twoDigitFlag = group == 0;
            }
        }
        return minus + output;
    }
    
    /**
     * Converts a {@code long} value into a string of words using the specified
     * language. Determines the scale based on the language.
     * 
     * @param number   the long value
     * @param language the language to use
     * @return a string of words representing {@code number}
     * @see #toWords(long)
     */
    public static String toWords(long number, String language) {
        return new NumberString(language).toWords(number);
    }
    
    /**
     * Converts a {@code long} value into a string of words using the specified
     * language and scale.
     * 
     * @param number   the long value
     * @param language the language to use
     * @param scale    the scale to use
     * @return a string of words representing number
     * @see #toWords(long)
     */
    public static String toWords(long number, String language, int scale) {
        return new NumberString(language, scale).toWords(number);
    }
    
    /**
     * Converts a {@code double} value into a string of words. Tentative
     * <ul>
     * <li>English: Supported</li>
     * <li>French: TODO: plurals</li>
     * <li>German: TODO: capitalisation, spaces</li>
     * </ul>
     * 
     * @param number the double value
     * @return a string of words representing number
     */
    public String toWords(double number) {
        String output = "", minus = "", temp = "", groupName = "";
        int curSec = 0, groupLength = 0, group = 0;
        boolean twoDigitFlag = false;
        
        if (number == 0) return (en ? "zero" : fr ? "zéro" : de ? "null" :"");
        if (number < 0) {
            minus = (en || de ? "minus " : fr ? "moins " :"");
            number = -number;
        }
        if (Double.isInfinite(number)) return minus + (en ? "infinity" : fr ? "infinité" : de ? "<translate:de:infinity>" :""); // TODO: German
        
        String manStr, expStr;
        long mantissa, exponent; // exponent is in decimal
        
        temp = Double.toString(number);
        if (temp.matches(".*E.*")) {
            manStr = temp.split("E")[0];
            expStr = temp.split("E")[1];
        } else {
            manStr = temp;
            expStr = "0";
        }
        mantissa = Long.parseLong(manStr.replace(".",""));
        exponent = Long.parseLong(expStr) + (long)(1-manStr.length()+manStr.indexOf("."));
        while (mantissa % 10 == 0) {
            mantissa /= 10;
            exponent += 1;
        }
        
        group = (int)exponent;
        if (group == 1) {
            mantissa *= 10;
            group = 0;
        } else if (group > 3) {
            if (lsnm && group > 6) {
                mantissa *= (int)Math.pow(10, group % 6);
                group -= group % 6;
            } else {
                mantissa *= (int)Math.pow(10, group % 3);
                group -= group % 3;
            }
        }
        while (mantissa > 0) {
            if (group < 0) groupLength = 1;
            else if (group == 0) groupLength = 2;
            else if (group == 2) groupLength = 1;
            else groupLength = (lsnm ? 6 : 3);
            
            groupName = getGroupName(group);
            
            curSec = (int)(mantissa % (int)Math.pow(10, groupLength));
            if (curSec > 0 || group < 0) {
                temp = getWordsPart(curSec) + groupName;
                if (output.equals("")) {
                    output = temp;
                } else if (twoDigitFlag) {
                    output = temp + (en ? " and " : fr ? " " : de ? "" :"") + output;
                } else if (group <= 0) {
                    output = temp + (en || fr || de ? " " :"") + output;
                } else {
                    output = temp + (en || fr ? ", " : de ? "" :"") + output;
                }
                twoDigitFlag = group == 0;
            }
            if (group == -1) output = (en ? "point " : fr ? "virgule " : de ? "komma " :"") + output;
            
            mantissa /= (int)Math.pow(10, groupLength);
            group += groupLength;
        }
        if (group == 0) {
            output = (en ? "zero " : fr ? "zéro " : de ? "null " :"") + output;
        } else if (group < 0) {
            output = (en ? "zero point <"+getWordsPart(-group)+" zeros> " :
                      fr ? "zéro virgule <"+getWordsPart(-group)+" zéros> " :
                      de ? "null komma <"+getWordsPart(-group)+" Nullen> " :"") + output;
        }
        
        return minus + output;
    }
    
    private String getWordsPart(int n) {
        int thousands = n / 1000;
        int hundreds = n / 100 % 10;
        int tensAndUnits = n % 100;
        int tens = n / 10 % 10;
        int units = n % 10;
        
        String output = "";
        
        if (thousands > 0){
            output = getWordsPart(thousands) + (en ? " thousand" : fr ? " mille" : de ? "tausend" :"");
            if (hundreds>0) output += (en || fr ? ", " : de ? "" :"") + getWordsPart(hundreds) + (en ? " hundred" : fr ? " cent" : de ? "hundert" :"");
            if (tensAndUnits>0) output += (en ? " and " : fr ? " " : de ? "" :"") + getWordsPart(tensAndUnits);
            
        } else if (hundreds > 0) {
            output = getWordsPart(hundreds) + (en ? " hundred" : fr ? " cent" : de ? "hundert" :"");
            if (tensAndUnits > 0) output += (en ? " and " : fr ? " " : de ? "" :"") + getWordsPart(tensAndUnits);
            
        } else if (tens > 1) {
            switch (tens) {
                case 2: output = (en ? "twenty"  : fr ? "vingt"        : de ? "zwanzig" :""); break;
                case 3: output = (en ? "thirty"  : fr ? "trente"       : de ? "dreißig" :""); break;
                case 4: output = (en ? "forty"   : fr ? "quarante"     : de ? "vierzig" :""); break;
                case 5: output = (en ? "fifty"   : fr ? "cinquante"    : de ? "fünfzig" :""); break;
                case 6: output = (en ? "sixty"   : fr ? "soixante"     : de ? "sechzig" :""); break;
                case 7: output = (en ? "seventy" : fr ? "soixante"     : de ? "siebzig" :"");
                    if (fr) units += 10;
                    break;
                case 8: output = (en ? "eighty"  : fr ? "quatre-vingt" : de ? "achtzig" :""); break;
                case 9: output = (en ? "ninety"  : fr ? "quatre-vingt" : de ? "neunzig" :"");
                    if (fr) units += 10;
                    break;
            }
            if (units > 0) {
                if (en) {
                    output += "-" + getWordsPart(units);
                } else if (fr) {
                    output += (units==1 ? "-et-" : "-") + getWordsPart(units);
                } else if (de) {
                    output = getWordsPart(units) + "und" + output;
                }
            }
            
        } else {
            switch (n) {
                case  1: return (en ? "one"       : fr ? "un"       : de ? "ein"      :"");
                case  2: return (en ? "two"       : fr ? "deux"     : de ? "zwei"     :"");
                case  3: return (en ? "three"     : fr ? "trois"    : de ? "drei"     :"");
                case  4: return (en ? "four"      : fr ? "quatre"   : de ? "vier"     :"");
                case  5: return (en ? "five"      : fr ? "cinq"     : de ? "fünf"     :"");
                case  6: return (en ? "six"       : fr ? "six"      : de ? "sechs"    :"");
                case  7: return (en ? "seven"     : fr ? "sept"     : de ? "sieben"   :"");
                case  8: return (en ? "eight"     : fr ? "huit"     : de ? "acht"     :"");
                case  9: return (en ? "nine"      : fr ? "neuf"     : de ? "neun"     :"");
                case 10: return (en ? "ten"       : fr ? "dix"      : de ? "zehn"     :"");
                case 11: return (en ? "eleven"    : fr ? "onze"     : de ? "elf"      :"");
                case 12: return (en ? "twelve"    : fr ? "douze"    : de ? "zwölf"    :"");
                case 13: return (en ? "thirteen"  : fr ? "treize"   : de ? "dreizehn" :"");
                case 14: return (en ? "fourteen"  : fr ? "quatorze" : de ? "vierzehn" :"");
                case 15: return (en ? "fifteen"   : fr ? "quinze"   : de ? "fünfzehn" :"");
                case 16: return (en ? "sixteen"   : fr ? "seize"    : de ? "sechzehn" :"");
                case 17: return (en ? "seventeen" : fr ? "dix-sept" : de ? "siebzehn" :"");
                case 18: return (en ? "eighteen"  : fr ? "dix-huit" : de ? "achtzehn" :"");
                case 19: return (en ? "nineteen"  : fr ? "dix-neuf" : de ? "neunzehn" :"");
            }
            return (en ? "zero" : fr ? "zéro" : de ? "null" :"");
        }
        return output;
    }
    
    private String getGroupName(int powerOfTen) {
        if (powerOfTen <= 0) return "";
        switch (powerOfTen) {
            case 2: return (en ? " hundred" : fr ? " cent" : de ? "hundert" :"");
            case 3: return (en ? " thousand" : fr ? " mille" : de ? "tausend" :"");
        }
        StringBuilder result = new StringBuilder(en || fr ? " " : de ? "" :"");
        result.append(getLatinPrefix(ss ? powerOfTen/3-1 : powerOfTen/6)).append("illi");
        if (lsm && (powerOfTen % 6 == 3))
            result.append(en ? "ard" : fr ? "ards" : de ? "arden" :"");
        else
            result.append(en ? "on" : fr ? "ons" : de ? "onen" :"");
        return result.toString();
    }
    
    private String getLatinPrefix(int p) {
        int hundreds = p / 100;
        int tens = p / 10 % 10;
        int units = p % 10;
        String output = "";
        
        if (hundreds > 0) return "<" + p + ">"; // Do something about this.
        switch (tens) {
            case 0:
            switch (units) {
                case 0: return "";
                case 1: return "m";
                case 2: return "b";
                case 3: return "tr";
                case 4: return "quadr";
                case 5: return "quint";
                case 6: return "sext";
                case 7: return "sept";
                case 8: return "oct";
                case 9: return "non";
            }
            case 1:
            switch (units) {
                case 0: return "dec";
                case 1: return "undec";
                case 2: return "dodec";
                case 3: return "tredec";
                case 4: return "quattourdec";
                case 5: return "quindec";
                case 6: return "sexdec";
                case 7: return "septendec";
                case 8: return "octodec";
                case 9: return "novemdec";
            }
            case 2: output = "vigint";
            case 3: output = "trigint";
            case 4: output = "quadragint";
            case 5: output = "quinquagint";
            case 6: output = "sexagint";
            case 7: output = "septuagint";
            case 8: output = "octogint";
            case 9: output = "nonagint";
        }
        switch (units) {
            case 1: output = "un" + output; break;
            case 2: output = "duo" + output; break;
            case 3: output = "tres" + output; break;
            case 4: output = "quattuor" + output; break;
            case 5: output = "quinqua" + output; break;
            case 6: output = "ses" + output; break;
            case 7: output = "septe"+(tens==2?"m":"n") + output; break;
            case 8: output = "octo" + output; break;
            case 9: output = "nove"+(tens==2?"m":"n") + output; break;
        }
        return output;
    }
    
    /**
     * Converts a string of words into a {@code long} value.
     * IN PROGRESS. ENGLISH ONLY.
     * 
     * @param text a string of words
     * @return the long value of the string
     */
    public long longFromWords(String text) {
        text = text.toLowerCase().replace(" ", "");
        int minus = 1;
        long whole = 0, part6 = 0, part3 = 0;
        int i = 0;
        while (i < text.length()) {
            if (text.startsWith("and", i)) {
                i += 3;
            } else if (text.startsWith("minus", i)) {
                minus *= -1;
                i += 5;
                
            } else if (text.startsWith("ten", i)) {
                part3 += 10;
                i += 3;
            } else if (text.startsWith("eleven", i)) {
                part3 += 11;
                i += 6;
            } else if (text.startsWith("twelve", i)) {
                part3 += 12;
                i += 6;
            } else if (text.startsWith("thirteen", i)) {
                part3 += 13;
                i += 8;
            } else if (text.startsWith("fourteen", i)) {
                part3 += 14;
                i += 8;
            } else if (text.startsWith("fifteen", i)) {
                part3 += 15;
                i += 7;
            } else if (text.startsWith("sixteen", i)) {
                part3 += 16;
                i += 7;
            } else if (text.startsWith("seventeen", i)) {
                part3 += 17;
                i += 9;
            } else if (text.startsWith("eighteen", i)) {
                part3 += 18;
                i += 8;
            } else if (text.startsWith("nineteen", i)) {
                part3 += 19;
                i += 8;
            } else if (text.startsWith("twenty", i)) {
                part3 += 20;
                i += 6;
            } else if (text.startsWith("thirty", i)) {
                part3 += 30;
                i += 6;
            } else if (text.startsWith("forty", i)) {
                part3 += 40;
                i += 5;
            } else if (text.startsWith("fifty", i)) {
                part3 += 50;
                i += 5;
            } else if (text.startsWith("sixty", i)) {
                part3 += 60;
                i += 5;
            } else if (text.startsWith("seventy", i)) {
                part3 += 70;
                i += 7;
            } else if (text.startsWith("eighty", i)) {
                part3 += 80;
                i += 6;
            } else if (text.startsWith("ninety", i)) {
                part3 += 90;
                i += 6;
            } else if (text.startsWith("one", i)) {
                part3 += 1;
                i += 3;
            } else if (text.startsWith("two", i)) {
                part3 += 2;
                i += 3;
            } else if (text.startsWith("three", i)) {
                part3 += 3;
                i += 5;
            } else if (text.startsWith("four", i)) {
                part3 += 4;
                i += 4;
            } else if (text.startsWith("five", i)) {
                part3 += 5;
                i += 4;
            } else if (text.startsWith("six", i)) {
                part3 += 6;
                i += 3;
            } else if (text.startsWith("seven", i)) {
                part3 += 7;
                i += 5;
            } else if (text.startsWith("eight", i)) {
                part3 += 8;
                i += 5;
            } else if (text.startsWith("nine", i)) {
                part3 += 9;
                i += 4;
            } else if (text.startsWith("zero", i)) {
                part3 += 0;
                i += 4;
                
            } else if (text.startsWith("hundred", i)) {
                part3 *= 100;
                i += 7;
            } else if (text.startsWith("thousand", i)) {
                part6 = (part6 + part3) * 1000;
                part3 = 0;
                i += 8;
            } else {
                int dave = 0;
                if (text.startsWith("million", i)) {
                    dave = 2;
                    i += 7;
                } else if (text.startsWith("milliard", i)) {
                    dave = 3;
                    i += 8;
                } else if (text.startsWith("billion", i)) {
                    dave = ss ? 3 : 4;
                    i += 7;
                } else if (text.startsWith("billiard", i)) {
                    dave = 5;
                    i += 8;
                } else if (text.startsWith("trillion", i)) {
                    dave = ss ? 4 : 6;
                    i += 8;
                } else if (text.startsWith("trilliard", i)) {
                    dave = 7;
                    i += 9;
                } else if (text.startsWith("quadrillion", i)) {
                    dave = ss ? 5 : 8;
                    i += 11;
                } else if (text.startsWith("quadrilliard", i)) {
                    dave = 9;
                    i += 12;
                } else if (text.startsWith("quintillion", i)) {
                    dave = ss ? 6 : 10;
                    i += 11;
                }
                if (dave > 0) {
                    long mul = 1;
                    for (int j=0; j<dave; j++) mul*=1000;
                    part6 += part3;
                    whole = part6 == 0 ? whole * mul : whole + part6 * mul;
                    part3 = 0;
                    part6 = 0;
                } else i++; // No recognised word starting at this point
            }
        }
        return whole + part6 + part3;
    }
    
    /**
     * Converts a {@code long} value into a string of roman numerals. Warning:
     * do not use for values &gt;&gt;1000 since this can take a very long time
     * and generate a very long string.
     * 
     * @param number the long value
     * @return a string of letters representing {@code number}
     */
    public String toRomanNumerals(long number) {
        StringBuilder result = new StringBuilder();
        
        for (int i=0; i<(int)(number/1000); i++) result.append("M");
        switch ((int)(number / 100) % 10) {
            case 1: result.append("C"); break;
            case 2: result.append("CC"); break;
            case 3: result.append("CCC"); break;
            case 4: result.append("CD"); break;
            case 5: result.append("D"); break;
            case 6: result.append("DC"); break;
            case 7: result.append("DCC"); break;
            case 8: result.append("DCCC"); break;
            case 9: result.append("CM"); break;
        }
        switch ((int)(number / 10) % 10) {
            case 1: result.append("X"); break;
            case 2: result.append("XX"); break;
            case 3: result.append("XXX"); break;
            case 4: result.append("XL"); break;
            case 5: result.append("L"); break;
            case 6: result.append("LX"); break;
            case 7: result.append("LXX"); break;
            case 8: result.append("LXXX"); break;
            case 9: result.append("XC"); break;
        }
        switch ((int)number % 10) {
            case 1: result.append("I"); break;
            case 2: result.append("II"); break;
            case 3: result.append("III"); break;
            case 4: result.append("IV"); break;
            case 5: result.append("V"); break;
            case 6: result.append("VI"); break;
            case 7: result.append("VII"); break;
            case 8: result.append("VIII"); break;
            case 9: result.append("IX"); break;
        }
        return result.toString();
    }
    
    /**
     * For testing. Creates a single {@code NumberString} and a number and
     * attempts to output it using each of the three methods. Run with the
     * {@code -h} option to list options.
     * 
     * @param args the arguments
     */
    public static void main(String[] args) {
        java.util.Random r = new java.util.Random();
        long n = r.nextLong();
        double d = 0.0;
        boolean isDouble = false;
        NumberString numberString = new NumberString();
        
        for (int i=0; i<args.length; i++) {
            if (args[i].equals("-n")) {
                i++;
                n = numberString.longFromDigits(args[i]);
                isDouble = false;
            } else if (args[i].equals("-d")) {
                i++;
                d = Double.parseDouble(args[i]);
                isDouble = true;
            } else if (args[i].equals("-p")) {
                i++;
                n = numberString.longFromWords(args[i]);
                isDouble = false;
            } else if (args[i].equals("-nr")) {
                n = r.nextLong();
                isDouble = false;
            } else if (args[i].equals("-nr")) {
                d = Double.longBitsToDouble(r.nextLong());
                isDouble = true;
            } else if (args[i].equals("-en")) {
                numberString.setLanguage(ENGLISH);
            } else if (args[i].equals("-fr")) {
                numberString.setLanguage(FRENCH);
            } else if (args[i].equals("-de")) {
                numberString.setLanguage(GERMAN);
            } else if (args[i].equals("-s")) {
                numberString.setScale(SHORT_SCALE);
            } else if (args[i].equals("-l")) {
                numberString.setScale(LONG_SCALE_NO_MILLIARD);
            } else if (args[i].equals("-m")) {
                numberString.setScale(LONG_SCALE_MILLIARD);
            } else if (args[i].equals("-h") || args[i].equals("--help")) {
                System.out.println("smithers.extras.NumberString -- converts numbers to and from strings.");
                System.out.println("Usage:  java smithers.extras.NumberString [options...]");
                System.out.println("Creates a number and attempts to output it as digits and words.");
                System.out.println("");
                System.out.println("Options: ");
                System.out.println("  -n n  Sets the number to the integer n.");
                System.out.println("  -d d  Sets the number to the real value d.");
                System.out.println("  -p s  Sets the number to the integer obtained by parsing the string s. Currently only valid if the language is English.");
                System.out.println("  -nr   Sets the number to a random integer.");
                System.out.println("  -dr   Sets the number to a random real value.");
                System.out.println("  -en   Sets the language to English.");
                System.out.println("  -fr   Sets the language to French/Francais.");
                System.out.println("  -de   Sets the language to German/Deutsch.");
                System.out.println("  -s    Sets the scale to the short scale.");
                System.out.println("  -l    Sets the scale to the long scale (no milliard).");
                System.out.println("  -m    Sets the scale to the long scale with milliard.");
                System.out.println("  -h");
                System.out.println("  --help");
                System.out.println("        Print this help and exit.");
                return;
            }
        }
        
        if (isDouble) {
            System.out.println(numberString.toDigits(d));
            System.out.println(numberString.toWords(d));
        } else {
            System.out.println(numberString.toDigits(n));
            System.out.println(numberString.toWords(n));
            if (n > 0 && n < 10000) System.out.println(numberString.toRomanNumerals(n));
        }
    }
    
}
