// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;

/**
 * Class to store a hanjie puzzle.
 */
public class Hanjie implements Puzzle {
    
    int height, width;
    boolean hasColour;
    Color[] colours;
    int[][] rowClues;
    int[][] rowColours;
    int[][] columnClues;
    int[][] columnColours;
    private boolean isMutable = true;
    
    //Private constructor used by factory methods.
    private Hanjie() {}
    
    /**
     * Constructs a new black and white puzzle of the specified size.
     * 
     * @param height the height of the puzzle
     * @param width  the width of the puzzle
     * @throws IllegalArgumentException if the width or height are negative
     */
    public Hanjie(int height, int width) {
        if (height<0 || width<0) throw new IllegalArgumentException("Height and width must be non-negative.");
        this.height = height;
        this.width = width;
        this.hasColour = false;
        this.rowClues = new int[height][];
        this.columnClues = new int[width][];
    }
    
    /**
     * Constructs a new colour puzzle of the specified size.
     * 
     * @param height  the height of the puzzle
     * @param width   the width of the puzzle
     * @param colours the colours, where {@code colours[0]} is the blank colour
     * @throws IllegalArgumentException if the width or height are negative
     */
    public Hanjie(int height, int width, Color[] colours) {
        if (height<0 || width<0) throw new IllegalArgumentException("Height and width must be non-negative.");
        this.height = height;
        this.width = width;
        this.hasColour = true;
        this.rowClues = new int[height][];
        this.rowColours = new int[height][];
        this.columnClues = new int[width][];
        this.columnColours = new int[width][];
    }
    
    /**
     * Constructs a new black and white puzzle from a grid of clues. The size
     * is inferred from the arrays. Each of the clue arrays should contain an
     * {@code int[]} for each row/column, containing the clue for that
     * row/column.
     * 
     * @param rowClues    the clues for the rows
     * @param columnClues the clues for the columns
     */
    public Hanjie(int[][] rowClues, int[][] columnClues) {
        this.height = rowClues.length;
        this.width = columnClues.length;
        this.hasColour = false;
        this.rowClues = rowClues;
        this.columnClues = columnClues;
    }
    
    /**
     * Constructs a new colour puzzle from a grid of clues. The size is
     * inferred from the arrays. The clue and colour arrays must be the same
     * size, and the colours must be positive integers. Each of the clue arrays
     * should contain an {@code int[]} for each row/column, containing the clue
     * for that row/column.
     * 
     * @param rowClues      the clues for the rows
     * @param rowColours    the colours for the row clues
     * @param columnClues   the clues for the colums
     * @param columnColours the colours for the colum clues
     * @param colours       the colours, where {@code colours[0]} is the blank
     *                      colour
     */
    public Hanjie(int[][] rowClues, int[][] rowColours, int[][] columnClues, int[][] columnColours, Color[] colours) {
        this.height = rowClues.length;
        this.width = columnClues.length;
        this.hasColour = true;
        this.rowClues = rowClues;
        this.rowColours = rowColours;
        this.columnClues = columnClues;
        this.columnColours = columnColours;
        this.colours = colours;
        if (rowColours.length != height || columnColours.length != width)
            throw new IllegalArgumentException("Sizes of clue and colour arrays must match.");
        for (int i=0; i<height; i++) {
            if (rowColours[i].length != rowClues[i].length)
                throw new IllegalArgumentException("Sizes of clue and colour arrays must match.");
        }
        for (int i=0; i<width; i++) {
            if (columnColours[i].length != columnClues[i].length)
                throw new IllegalArgumentException("Sizes of clue and colour arrays must match.");
        }
    }
    
    /**
     * Loads a hanjie puzzle from a file.
     * 
     * @param file the file to load from
     * @throws IOException           if one occurs
     * @throws PuzzleFormatException if the file is not in the right format
     * @see #save(java.io.File)
     */
    public static Hanjie load(File file) throws IOException {
        Hanjie result = new Hanjie();
        BufferedReader in = new BufferedReader(new FileReader(file));
        String line="";
        int lineNumber = 0;
        String[] temp;
        String[] colourNames=null;
        int numCols = 0;
        int i, j, k;
        
        try {
            line = in.readLine();
            lineNumber++;
            result.hasColour = false;
            if (line.charAt(0) == 'C' || line.charAt(0) == 'c') {
                result.hasColour = true;
                
                line = in.readLine();
                lineNumber++;
                temp = line.split(" +");
                if (temp.length %2 != 1) throw new RuntimeException(); //Will be replaced by correct Exception in catch block
                numCols = (temp.length+1)/2;
                result.colours = new Color[numCols];
                colourNames = new String[numCols];
                for (i=0; i<numCols; i++) {
                    result.colours[i] = new Color(Integer.parseInt(temp[2*i], 16));
                    if (i>0) colourNames[i] = temp[2*i-1];
                }
                line = in.readLine();
                lineNumber++;
            }
            
            temp = line.split(" +");
            result.height = Integer.parseInt(temp[0]);
            result.width = Integer.parseInt(temp[1]);
            result.rowClues = new int[result.height][];
            result.columnClues = new int[result.width][];
            if (result.hasColour) {
                result.rowColours = new int[result.height][];
                result.columnColours = new int[result.width][];
            }
            
            for (i=0; i<result.height; i++) {
                line = in.readLine();
                lineNumber++;
                temp = line.split(" +");
                result.rowClues[i] = new int[temp.length];
                if (result.hasColour) result.rowColours[i] = new int[temp.length];
                for (j=0; j<temp.length; j++) {
                    if (result.hasColour) {
                        for (k=1; k<numCols; k++) {
                            if (temp[j].startsWith(colourNames[k])) break;
                        }
                        if (k==numCols) throw new RuntimeException(); //Will be replaced by correct Exception in catch block
                        result.rowClues[i][j] = Integer.parseInt(temp[j].substring(colourNames[k].length()));
                        result.rowColours[i][j] = k;
                    } else {
                        result.rowClues[i][j] = Integer.parseInt(temp[j]);
                    }
                }
            }
            
            for (i=0; i<result.width; i++) {
                line = in.readLine();
                lineNumber++;
                temp = line.split(" +");
                result.columnClues[i] = new int[temp.length];
                if (result.hasColour) result.columnColours[i] = new int[temp.length];
                for (j=0; j<temp.length; j++) {
                    if (result.hasColour) {
                        for (k=1; k<numCols; k++) {
                            if (temp[j].startsWith(colourNames[k])) break;
                        }
                        if (k==numCols) throw new RuntimeException(); //Will be replaced by correct Exception in catch block
                        result.columnClues[i][j] = Integer.parseInt(temp[j].substring(colourNames[k].length()));
                        result.columnColours[i][j] = k;
                    } else {
                        result.columnClues[i][j] = Integer.parseInt(temp[j]);
                    }
                }
            }
            
            return result;
        } catch (RuntimeException e) {
            throw new PuzzleFormatException(line, lineNumber);
        }
    }
    
    /**
     * Saves the puzzle to a file.
     * 
     * @param file the file to save to
     * @throws IOException if one occurs
     * @see #load(java.io.File)
     */
    public void save(File file) throws IOException {
        PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(file)));
        out.print(toString());
        out.flush();
        out.close();
    }
    
    public boolean isMutable() {
        return isMutable;
    }
    
    public Hanjie makeImmutableCopy() {
        Hanjie result = new Hanjie();
        result.height = height;
        result.width = width;
        result.hasColour = hasColour;
        result.colours = colours;
        result.rowClues = Arrays.copyOf(rowClues, rowClues.length);
        result.columnClues = Arrays.copyOf(columnClues, columnClues.length);
        if (hasColour) {
            result.rowColours = Arrays.copyOf(rowColours, rowColours.length);
            result.columnColours = Arrays.copyOf(columnColours, columnColours.length);
        }
        result.isMutable = false;
        return result;
    }
    
    /**
     * Gets the height of the puzzle.
     * 
     * @return the height of the puzzle
     */
    public int getHeight() {
        return height;
    }
    
    /**
     * Gets the width of the puzzle.
     * 
     * @return the width of the puzzle
     */
    public int getWidth() {
        return width;
    }
    
    /**
     * Gets the clue for the specified row.
     * 
     * @param row the row index (starting from 0)
     * @return the clue for the {@code row}th row
     */
    public int[] getRowClue(int row) {
        if (row < 0 || row >= height) throw new IllegalArgumentException("Invalid row: " + row + ".");
        return rowClues[row];
    }
    
    /**
     * Gets the colour information for the clue for the specified row. Throws
     * an exception if the puzzle does not have colour.
     * 
     * @param row the row index (starting from 0)
     * @return the colours for the clue for the {@code row}th row
     */
    public int[] getRowColour(int row) {
        if (!hasColour) throw new IllegalArgumentException("Puzzle does not have colour.");
        if (row < 0 || row >= height) throw new IllegalArgumentException("Invalid row: " + row + ".");
        return rowColours[row];
    }
    
    /**
     * Gets the clue for the specified column.
     * 
     * @param column the column index (starting from 0)
     * @return the clue for the {@code column}th column
     */
    public int[] getColumnClue(int column) {
        if (column < 0 || column >= width) throw new IllegalArgumentException("Invalid column: " + column + ".");
        return columnClues[column];
    }
    
    /**
     * Gets the colour information for the clue for the specified column.
     * Throws an exception if the puzzle does not have colour.
     * 
     * @param column the column index (starting from 0)
     * @return the colours for the clue for the {@code column}th column
     */
    public int[] getColumnColour(int column) {
        if (!hasColour) throw new IllegalArgumentException("Puzzle does not have colour.");
        if (column < 0 || column >= width) throw new IllegalArgumentException("Invalid column: " + column + ".");
        return columnColours[column];
    }
    
    /**
     * Gets whether or not this puzzle has colour.
     * 
     * @return {@code true} iff this is a colour puzzle
     */
    public boolean hasColour() {
        return hasColour;
    }
    
    /**
     * Gets the number of colours.
     * 
     * @return the number of colours, or 2 if this puzzle is black and white
     */
    public int getColourCount() {
        if (!hasColour) return 2;
        else return colours.length;
    }
    
    /**
     * Gets the colour array.
     * 
     * @return the colours, or {@code null} if this puzzle is black and white
     */
    public Color[] getColours() {
        if (!hasColour) return null;
        return colours;
    }
    
    /**
     * Adds a clue to a black and white puzzle. The clue is an {@code int[]}
     * specifying the clue for an enitire row/column. To add a clue to a colour
     * puzzle, use {@code addClue(boolean, int, int[], int[])}.
     * 
     * @param isRow {@code true} if the clue is for a row, {@code false} for a
     *              column
     * @param index the position of the clue
     * @param clue  the clue to add
     * @see #addClue(boolean, int, int[], int[])
     */
    public void addClue(boolean isRow, int index, int[] clue) {
        if (!isMutable) throw new ImmutablePuzzleException();
        if (hasColour) throw new IllegalArgumentException("Attempting to add black and white clue to a colour puzzle.");
        if (isRow) {
            if (index < 0 || index >= height)
                throw new IllegalArgumentException("Invalid row: " + index + ".");
            rowClues[index] = clue;
        } else {
            if (index < 0 || index >= width)
                throw new IllegalArgumentException("Invalid column: " + index + ".");
            columnClues[index] = clue;
        }
    }
    
    /**
     * Adds a clue to a colour puzzle. The clue is 2 {@code int[]}s specifying
     * the clue for an enitire row/column. The {@code clue} and {@code colour}
     * arguments should have the same length. To add a clue to a black and
     * white puzzle, use {@code addClue(boolean, int, int[])}.
     * 
     * @param isRow  {@code true} if the clue is for a row, {@code false} for a
     *               column
     * @param index  the position of the clue
     * @param clue   the clue to add
     * @param colour the colours of the clue
     * @see #addClue(boolean, int, int[])
     */
    public void addClue(boolean isRow, int index, int[] clue, int[] colour) {
        if (!isMutable) throw new ImmutablePuzzleException();
        if (!hasColour) throw new IllegalArgumentException("Attempting to add colour clue to a black and white puzzle.");
        if (colour.length != clue.length)
            throw new IllegalArgumentException("Sizes of clue and colour arrays must match.");
        if (isRow) {
            if (index < 0 || index >= height)
                throw new IllegalArgumentException("Invalid row: " + index + ".");
            rowClues[index] = clue;
            rowColours[index] = colour;
        } else {
            if (index < 0 || index >= width)
                throw new IllegalArgumentException("Invalid column: " + index + ".");
            columnClues[index] = clue;
            columnColours[index] = colour;
        }
    }
    
    /**
     * Generates a {@code String} representing the puzzle's clues.
     * 
     * @return a {@code String} representing the puzzle
     */
    public String toString() {
        StringBuilder result = new StringBuilder();
        if (hasColour) {
            result.append("C\n").append(colours[0]);
            for (int i=1; i<colours.length; i++) {
                result.append(' ').append((char)('a'+i-1)).append(' ').append(Integer.toHexString(colours[i].getRGB()));
            }
            result.append('\n');
        }
        result.append(height).append(' ').append(width).append('\n');
        for (int i=0; i<height; i++) {
            for (int j=0; j<rowClues[i].length; j++) {
                if (j>0) result.append(' ');
                if (hasColour) result.append((char)('a'+rowColours[i][j]-1));
                result.append(rowClues[i][j]);
            }
            result.append('\n');
        }
        for (int i=0; i<width; i++) {
            for (int j=0; j<columnClues[i].length; j++) {
                if (j>0) result.append(' ');
                if (hasColour) result.append((char)('a'+columnColours[i][j]-1));
                result.append(columnClues[i][j]);
            }
            result.append('\n');
        }
        return result.toString();
    }
    
    public static void main(String[] args) throws IOException {
        if (args.length == 0 || args[0].equals("-h") || args[0].equals("--help")) {
            System.out.println("smithers.logicpuzzles.Hanjie -- Graphical solver of hanjie puzzles.");
            System.out.println("Usage:");
            System.out.println("        java smithers.logicpuzzles.Hanjie -h|--help");
            System.out.println("        java smithers.logicpuzzles.Hanjie puzzle");
            System.out.println("");
            System.out.println("If -h or --help is given, display this help and exit.");
            System.out.println("Otherwise puzzle must be the path to a hanjie puzzle file (in the format output by smithers.logicpuzzles.HanjieBuilder).");
            System.out.println("In this case, the puzzle is displayed on the screen. " +
                               "Click outside the puzzle to solve it; left-click to advance one step, middle-click to advance one pass or right-click to solve as completely as possible.");
            return;
        }
        smithers.util.Launcher.launch(new PuzzleComponent(new HanjieGraphics(new HanjieSolver(load(new File(args[0]))))), "Hanjie by Smithers", 500, 500);
    }
    
}
