// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;

public class MosaicGraphics implements PuzzleGraphics {
    
    public static final Color COLOUR_UNKNOWN = Color.LIGHT_GRAY;
    public static final Color COLOUR_BORDERS = Color.DARK_GRAY;
    public static final Color COLOUR_CLEAR = Color.WHITE;
    public static final Color COLOUR_FILLED = Color.BLACK;
    
    private Mosaic puzzle;
    private MosaicState state;
    private boolean hasSolver;
    private MosaicSolver solver;
    private int height, width;
    private Rectangle drawn = null;
    
    public MosaicGraphics(MosaicState state) {
        this.state = state;
        puzzle = state.getPuzzle();
        hasSolver = false;
        height = puzzle.getHeight();
        width = puzzle.getWidth();
    }
    
    public MosaicGraphics(MosaicSolver solver) {
        this.solver = solver;
        state = solver.getState();
        puzzle = solver.getPuzzle();
        hasSolver = true;
        height = puzzle.getHeight();
        width = puzzle.getWidth();
    }
    
    public Rectangle draw(Graphics g, Rectangle r) {
        int gridWidth = (r.width - 1) / width;
        int gridHeight = (r.height - 1) / height;
        int gridSize = gridWidth < gridHeight ? gridWidth : gridHeight;
        int boxSize = gridSize - 1;
        int typeSize, textRelX=0, textRelY=0;
        int x = r.x + (r.width - gridSize * width - 1)/2;
        int y = r.y + (r.height - gridSize * height - 1)/2;
        Rectangle cb = g.getClipBounds();
        drawn = new Rectangle(x, y, gridSize * width + 1, gridSize * height + 1);
        
        if (!g.hitClip(drawn.x, drawn.y, drawn.width, drawn.height)) return drawn;
        
        for (typeSize = 24; typeSize >= 0; typeSize -= 1) {
            g.setFont(new Font("Dialog", Font.PLAIN, typeSize));
            FontMetrics fm = g.getFontMetrics();
            int textWidth = fm.charWidth('0');
            int textAscent = fm.getAscent();
            int textDescent = fm.getDescent();
            
            if (textWidth < boxSize && textAscent + textDescent < boxSize) {
                textRelX = boxSize/2 - textWidth/2;
                textRelY = boxSize/2 + textAscent/2;
                break;
            }
        }
        
        int iMin = (cb.y - y) / gridSize;
        int iMax = (cb.y + cb.height - y - 2) / gridSize + 1;
        int jMin = (cb.x - x) / gridSize;
        int jMax = (cb.x + cb.width - x - 2) / gridSize + 1;
        if (iMin < 0) iMin = 0;
        if (iMax > height) iMax = height;
        if (jMin < 0) jMin = 0;
        if (jMax > width) jMax = width;
        
        g.setColor(COLOUR_BORDERS);
        g.fillRect(drawn.x, drawn.y, drawn.width, drawn.height);
        for (int i=iMin; i<iMax; i++) for (int j=jMin; j<jMax; j++) {
            int cell = state.getState(i, j);
            int clue = puzzle.getClue(i, j);
            g.setColor(cell == 0 ? COLOUR_CLEAR : cell == 1 ? COLOUR_FILLED : COLOUR_UNKNOWN);
            g.fillRect(x + j * gridSize + 1, y + i * gridSize + 1, boxSize, boxSize);
            if (clue >= 0 && typeSize >= 0) {
                g.setColor(cell == 1 ? Color.WHITE : Color.BLACK);
                g.drawString(Integer.toString(clue), x + j * gridSize + 1 + textRelX, y + i * gridSize + 1 + textRelY);
            }
        }
        
        int[] lastChange = solver.getLastChange();
        if (lastChange != null) {
            g.setColor(Color.RED);
            for (int k=0; k<lastChange.length/2; k++) {
                int i = lastChange[2*k];
                int j = lastChange[2*k+1];
                int x0 = x + (j==0 ? 0 : j-1) * gridSize;
                int x1 = x + (j==width-1 ? width : j+2) * gridSize;
                int y0 = y + (i==0 ? 0 : i-1) * gridSize;
                int y1 = y + (i==height-1 ? height : i+2) * gridSize;
                if (g.hitClip(x0-1, y0-1, x1-x0+3, y1-y0+3))
                    for (int d=-1; d<2; d++)
                        g.drawRect(x0 + d, y0 + d, x1 - x0 - 2*d, y1 - y0 - 2*d);
            }
        }
        
        return drawn;
    }
    
//     private Rectangle getCellBounds(int i, int j) {
//         int gridSize = (drawn.height - 1) / height;
//         return new Rectangle(drawn.x + j * gridSize + 1, drawn.y + i * gridSize + 1, gridSize - 1, gridSize - 1);
//     }
    
    private Rectangle getBoxBounds(int i, int j) {
        int gridSize = (drawn.height - 1) / height;
        int x0 = drawn.x + (j==0 ? 0 : j-1) * gridSize;
        int x1 = drawn.x + (j==width-1 ? width : j+2) * gridSize;
        int y0 = drawn.y + (i==0 ? 0 : i-1) * gridSize;
        int y1 = drawn.y + (i==height-1 ? height : i+2) * gridSize;
        return new Rectangle(x0-1, y0-1, x1-x0+3, y1-y0+3);
    }
    
    public void mouseClicked(MouseEvent e) {
        if (hasSolver) {
            if (e.getButton() == MouseEvent.BUTTON1) {
                int[] lastChange = solver.getLastChange();
                solver.step();
                int[] change = solver.getLastChange();
                if (drawn == null) {
                    e.getComponent().repaint();
                } else {
                    Rectangle toPaint;
                    if (lastChange != null)
                        for (int i=0; i<lastChange.length/2; i++)
                        {
                            toPaint = getBoxBounds(lastChange[2*i], lastChange[2*i+1]);
                            e.getComponent().repaint(toPaint.x, toPaint.y, toPaint.width, toPaint.height);
                        }
                    if (change != null)
                        for (int i=0; i<change.length/2; i++)
                        {
                            toPaint = getBoxBounds(change[2*i], change[2*i+1]);
                            e.getComponent().repaint(toPaint.x, toPaint.y, toPaint.width, toPaint.height);
                        }
                }
                lastChange = change;
            } else if (e.getButton() == MouseEvent.BUTTON2) {
                solver.pass();
                e.getComponent().repaint();
            } else if (e.getButton() == MouseEvent.BUTTON3) {
                solver.solve();
                e.getComponent().repaint();
            }
        }
    }
    
    /* Specified by MouseListener */
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}
    
}
