// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.util.Random;

/**
 * Solves mosaic puzzles. The method employed here uses two modes: simple logic
 * and advanced logic. Simple logic attempts to solve the puzzle only looking
 * at one clue at a time, whereas advanced logic looks at up to two clues at
 * once.
 */
public class MosaicSolver extends AbstractPuzzleSolver<Mosaic, MosaicState> {
    
    private int height, width;
    
    private boolean advancedLogic;
    private int[] clueI;
    private int[] clueJ;
    private int[] clueC;
    private int clueIndex;
    private int passSolved;
    private int[] lastChange;
    private int[][] myState;
    
    /**
     * Creates a new solver for the specified puzzle.
     * 
     * @param puzzle the puzzle to solve
     */
    public MosaicSolver(Mosaic puzzle) {
        this(puzzle, true);
    }
    
    /**
     * Creates a new solver to work from the specified state.
     * 
     * @param state the puzzle state to work from
     */
    public MosaicSolver(MosaicState state) {
        this(state, true);
    }
    
    /**
     * Creates a new solver for the specified puzzle.
     * 
     * @param puzzle   the puzzle to solve
     * @param advanced whether or not to use advanced logic
     */
    public MosaicSolver(Mosaic puzzle, boolean advanced) {
        super(puzzle);
        height = puzzle.height;
        width = puzzle.width;
        this.advancedLogic = advanced;
        start();
    }
    
    /**
     * Creates a new solver to work from the specified state.
     * 
     * @param state    the puzzle state to work from
     * @param advanced whether or not to use advanced logic
     */
    public MosaicSolver(MosaicState state, boolean advanced) {
        super(state);
        height = puzzle.height;
        width = puzzle.width;
        this.advancedLogic = advanced;
        start();
    }
    
    public MosaicState makeState(Mosaic puzzle) {
        return new MosaicState(puzzle);
    }
    
    public void start() {
        int i, j, k;
        int clue;
        
        clueI = new int[height*width+1];
        clueJ = new int[height*width+1];
        clueC = new int[height*width+1];
        myState = new int[height][width];
        
        k=0;
        for (i=0; i<height; i++) for (j=0; j<width; j++) {
            clue = puzzle.getClue(i, j);
            if (clue >= 0) {
                clueI[k] = i;
                clueJ[k] = j;
                clueC[k] = clue;
                k++;
            }
            myState[i][j] = state.getState(i, j);
        }
        
        // A bit more interesting.
        // Also means MosaicMaker will generate better puzzles.
        Random rand = new Random();
        for (i=k-1; i>0; i--) {
            j = rand.nextInt(i+1);
            clueI[k] = clueI[i];
            clueI[i] = clueI[j];
            clueI[j] = clueI[k];
            clueJ[k] = clueJ[i];
            clueJ[i] = clueJ[j];
            clueJ[j] = clueJ[k];
            clueC[k] = clueC[i];
            clueC[i] = clueC[j];
            clueC[j] = clueC[k];
        }
        clueC[k] = -1;
        clueIndex = 0;
        passSolved = 0;
    }
    
    /**
     * Finds the last change that was made to the puzzle. If the last call to
     * {@code step()} or {@code pass()} resulted in no change, this returns
     * {@code null}.
     * 
     * @return an array containing the row and column indices of the clues used
     */
    public int[] getLastChange() {
        return lastChange;
    }
    
    public boolean nextStep() {
        clueIndex++;
        if (clueC[clueIndex] == -1) {
            clueC[clueIndex-passSolved] = -1;
            clueIndex = 0;
            passSolved = 0;
            return true;
        }
        return false;
    }
    
    /**
     * Solves the current clue as much as possible. This will first check if
     * anything can be deduced from the current clue and then move on (if
     * advanced logic is enabled) to deductions from this clue and another clue
     * tested earlier in the pass.
     * 
     * @return {@code 2} if a change was made, {@code 0} otherwise
     */
    public int solveStep() {
        lastChange = null;
        if (clueC[clueIndex] == -1) return 0;
        //If a clue has been solved, shift other clues over it
        if (passSolved>0) {
            clueI[clueIndex-passSolved] = clueI[clueIndex];
            clueJ[clueIndex-passSolved] = clueJ[clueIndex];
            clueC[clueIndex-passSolved] = clueC[clueIndex];
        }
        
        int[] nb = neighbours(clueI[clueIndex], clueJ[clueIndex]);
        if (clueC[clueIndex] == nb[2]) {
            passSolved++;
            if (fillNeighbours(clueI[clueIndex], clueJ[clueIndex], 0)) {
                lastChange = new int[2];
                lastChange[0] = clueI[clueIndex];
                lastChange[1] = clueJ[clueIndex];
            }
        } else if (clueC[clueIndex] == nb[0] + nb[2]) {
            passSolved++;
            if (fillNeighbours(clueI[clueIndex], clueJ[clueIndex], 1)) {
                lastChange = new int[2];
                lastChange[0] = clueI[clueIndex];
                lastChange[1] = clueJ[clueIndex];
            }
        } else if (advancedLogic) {
            for (int clue2=0; clue2<clueIndex-passSolved; clue2++) {
                if (clueI[clueIndex]-clueI[clue2]>2 || clueI[clue2]-clueI[clueIndex]>2 || clueJ[clueIndex]-clueJ[clue2]>2 || clueJ[clue2]-clueJ[clueIndex]>2) continue;
                int[][] nb2 = neighbours(clueI[clueIndex], clueJ[clueIndex], clueI[clue2], clueJ[clue2]);
                if (clueC[clueIndex] - nb2[0][0] - nb2[0][2] == clueC[clue2] - nb2[2][2]) {
                    passSolved++;
                    if (fillNeighbours(clueI[clue2], clueJ[clue2], clueI[clueIndex], clueJ[clueIndex])) {
                        lastChange = new int[4];
                        lastChange[0] = clueI[clueIndex];
                        lastChange[1] = clueJ[clueIndex];
                        lastChange[2] = clueI[clue2];
                        lastChange[3] = clueJ[clue2];
                    }
                    break;
                } else if (clueC[clue2] - nb2[2][0] - nb2[2][2] == clueC[clueIndex] - nb2[0][2]) {
                    passSolved++;
                    if (fillNeighbours(clueI[clueIndex], clueJ[clueIndex], clueI[clue2], clueJ[clue2])) {
                        lastChange = new int[4];
                        lastChange[0] = clueI[clueIndex];
                        lastChange[1] = clueJ[clueIndex];
                        lastChange[2] = clueI[clue2];
                        lastChange[3] = clueJ[clue2];
                    }
                    break;
                }
            }
        }
        
        return lastChange == null ? 0 : 2;
    }
    
    private boolean updateState(int i, int j, int newState) {
        if (myState[i][j] == -1) {
            myState[i][j] = newState;
            state.setState(i, j, newState);
            return true;
        }
        return false;
    }
    
    // Counts the neighbours in each state (including the cell itself).
    // <result>[a] = # of neighbours in state a-1
    private int[] neighbours(int i0, int j0) {
        int[] result = new int[3];
        for (int i = i0>0 ? i0-1 : 0; i<=i0+1 && i<height; i++)
            for (int j = j0>0 ? j0-1 : 0; j<=j0+1 && j<width; j++)
                result[myState[i][j]+1]++;
        return result;
    }
    
    // Counts the neighbours and common neighbours in each state.
    // [0][] = neighbours of i0,j0 only
    // [1][] = common neighbours
    // [2][] = neighbours of i1,j1 only
    private int[][] neighbours(int i0, int j0, int i1, int j1) {
        int[][] result = new int[3][3];
        int iMin = i0<i1 ? i0 : i1;
        int jMin = j0<j1 ? j0 : j1;
        int iMax = i0>i1 ? i0 : i1;
        int jMax = j0>j1 ? j0 : j1;
        result[0] = neighbours(i0, j0);
        result[2] = neighbours(i1, j1);
        for (int i = iMax>0 ? iMax-1 : 0; i<=iMin+1 && i<height; i++) {
            for (int j = jMax>0 ? jMax-1 : 0; j<=jMin+1 && j<width; j++) {
                result[0][myState[i][j]+1]--;
                result[1][myState[i][j]+1]++;
                result[2][myState[i][j]+1]--;
            }
        }
        return result;
    }
    
    // Fills remaining neighbours with a specific state.
    private boolean fillNeighbours(int i0, int j0, int newState) {
        boolean result = false;
        for (int i = i0>0 ? i0-1 : 0; i<=i0+1 && i<height; i++)
            for (int j = j0>0 ? j0-1 : 0; j<=j0+1 && j<width; j++)
                result |= updateState(i, j, newState);
        return result;
    }
    
    // Fills remaining non-common neighbours with different states.
    // Fills neighbours of i0,j0 only with 0.
    // Fills neighbours of i1,j1 only with 1.
    private boolean fillNeighbours(int i0, int j0, int i1, int j1) {
        boolean result = false;
        for (int i = i0>0 ? i0-1 : 0; i<=i0+1 && i<height; i++)
            for (int j = j0>0 ? j0-1 : 0; j<=j0+1 && j<width; j++)
                if (i<i1-1 || i>i1+1 || j<j1-1 || j>j1+1)
                    result |= updateState(i, j, 0);
        for (int i = i1>0 ? i1-1 : 0; i<=i1+1 && i<height; i++)
            for (int j = j1>0 ? j1-1 : 0; j<=j1+1 && j<width; j++)
                if (i<i0-1 || i>i0+1 || j<j0-1 || j>j0+1)
                    result |= updateState(i, j, 1);
        return result;
    }
    
}
