// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

/**
 * Represents a slitherlink puzzle.
 */
public class Slitherlink implements Puzzle {
    
    int height, width;
    int[][] clues;
    private boolean isMutable = true;
    
    //Private constructor used by factory methods.
    private Slitherlink() {}
    
    /**
     * Constructs a blank puzzle of the specified size.
     * 
     * @param height the height of the puzzle
     * @param width  the width of the puzzle
     * @throws IllegalArgumentException if the width or height are negative
     */
    public Slitherlink(int height, int width) {
        if (height<0 || width<0) throw new IllegalArgumentException("Height and width must be non-negative.");
        this.height = height;
        this.width = width;
        this.clues = new int[height][width];
        for (int i=0; i<height; i++) for (int j=0; j<width; j++)
            clues[i][j] = -1;
    }
    
    /**
     * Constructs a puzzle from a grid of clues. The size is inferred from the
     * array.
     * 
     * @param clues an array of the clues, with -1 representing no clue
     * @throws IllegalArgumentException if any of the clues are out of range
     */
    public Slitherlink(int[][] clues) {
        this.height = clues.length;
        this.width = clues[0].length;
        this.clues = new int[height][width];
        for (int i=0; i<height; i++) for (int j=0; j<width; j++) {
            if (clues[i][j] < -1 || clues[i][j] > 9)
                throw new IllegalArgumentException("Array of clues is invalid.");
            this.clues[i][j] = clues[i][j];
        }
    }
    
    /**
     * Loads a puzzle from a file.
     * 
     * @param file the file to load from
     * @throws IOException if one occurs
     */
    public static Slitherlink load(File file) throws IOException {
        Slitherlink result = new Slitherlink();
        BufferedReader in = new BufferedReader(new FileReader(file));
        String input, temp;
        
        input = in.readLine();
        result.height = Integer.parseInt(input.split(" +")[0]);
        result.width = Integer.parseInt(input.split(" +")[1]);
        result.clues = new int[result.height][result.width];
        
        for (int i=0; i<result.height; i++) {
            input = in.readLine();
            for (int j=0; j<result.width; j++) {
                temp = input.substring(j,j+1);
                result.clues[i][j] = temp.equals(" ") ? -1 : Integer.parseInt(temp);
            }
        }
        return result;
    }
    
    /**
     * Saves the puzzle to a file.
     * 
     * @param file the file to save to
     * @throws IOException if one occurs
     */
    public void save(File file) throws IOException {
        PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(file)));
        out.print(toString());
        out.flush();
        out.close();
    }
    
    public boolean isMutable() {
        return isMutable;
    }
    
    public Slitherlink makeImmutableCopy() {
        Slitherlink result = new Slitherlink();
        result.height = height;
        result.width = width;
        result.clues = new int[height][width];
        for (int i=0; i<height; i++) for (int j=0; j<width; j++)
            result.clues[i][j] = clues[i][j];
        return result;
    }
    
    /**
     * Gets the height of the puzzle.
     * 
     * @return the height of the puzzle
     */
    public int getHeight() {
        return height;
    }
    
    /**
     * Gets the width of the puzzle.
     * 
     * @return the width of the puzzle
     */
    public int getWidth() {
        return width;
    }
    
    /**
     * Gets the clue for the specified cell.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @return the clue in the cell in row {@code row}, column {@code column},
     *         or {@code -1} if the cell is blank
     */
    public int getClue(int row, int column) {
        if (row < 0 || column < 0 || row >= height || column >= width)
            throw new IllegalArgumentException("Invalid location: " + row + ", " + column + ".");
        return clues[row][column];
    }
    
    /**
     * Sets the clue for the specified cell.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @param clue   the clue to place in the cell, or {@code -1} to clear
     */
    public void setClue(int row, int column, int clue) {
        if (!isMutable) throw new ImmutablePuzzleException();
        if (row < 0 || column < 0 || row >= height || column >= width)
            throw new IllegalArgumentException("Invalid location: " + row + ", " + column + ".");
        if (clue < -1 || clue > 9)
            throw new IllegalArgumentException("Invalid clue: " + clue + ".");
        clues[row][column] = clue;
    }
    
    /**
     * Generates a {@code String} representing the puzzle's clues. Uses spaces
     * to represent no clue.
     * 
     * @return a {@code String} representing the puzzle
     */
    public String toString() {
        char[] result = new char[(width + 1) * height];
        for (int i=0; i<height; i++) {
            for (int j=0; j<width; j++) {
                result[i * (width + 1) + j] = clues[i][j] == -1 ? ' ' : (char) (clues[i][j] + '0');
            }
            result[i * (width + 1) + width] = '\n';
        }
        return height + " " + width + "\n" + new String(result);
    }
    
    public static void main(String[] args) throws IOException {
//         Slitherlink puzzle = load(new File(args[0]));
//         SlitherlinkState state = new SlitherlinkState(puzzle);
//         SlitherlinkSolver solver = new SlitherlinkSolver(state);
//         do System.out.println(state.toString());
//         while (solver.pass());
        smithers.util.Launcher.launch(new PuzzleComponent(new SlitherlinkGraphics(new SlitherlinkSolver(load(new File(args[0]))))), "Slitherlink by Smithers", 500, 500);
    }
    
}
