// Copyright 2010 Christopher Smith
// 
// This file is part of smithers.logicpuzzles.
// 
// smithers.logicpuzzles is free software: you can redistribute it and/or
// modify it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
// 
// smithers.logicpuzzles is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// smithers.logicpuzzles. If not, see <http://www.gnu.org/licenses/>.

package smithers.logicpuzzles;

/**
 * Represents a state of a slitherlink puzzle. Note: {@link #isSolved()} is not
 * implemented yet.
 */
public class SlitherlinkState implements PuzzleState<Slitherlink> {
    
    Slitherlink puzzle;
    int height, width;
    int[][] horizontal;
    int[][] vertical;
    
    /**
     * Constructs a state for the specified puzzle.
     * 
     * @param puzzle the puzzle that this is a state of
     */
    public SlitherlinkState(Slitherlink puzzle) {
        this.puzzle = puzzle.isMutable() ? puzzle.makeImmutableCopy() : puzzle;
        height = this.puzzle.height;
        width = this.puzzle.width;
        this.horizontal = new int[height+1][width];
        this.vertical = new int[height][width+1];
        for (int i=0; i<height+1; i++) for (int j=0; j<width; j++)
            horizontal[i][j] = -1;
        for (int i=0; i<height; i++) for (int j=0; j<width+1; j++)
            vertical[i][j] = -1;
    }
    
    public Slitherlink getPuzzle() {
        return puzzle;
    }
    
    /**
     * Gets the state of the specified horizontal wall.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @return 1 if the wall is present, 0 if not present, or -1 if unknown
     */
    public int getHorizontalState(int row, int column) {
        if (row < 0 || column < 0 || row > height || column >= width)
            throw new IllegalArgumentException("Invalid location for horizontal wall: " + row + ", " + column + ".");
        return horizontal[row][column];
    }
    
    /**
     * Gets the state of the specified vertical wall.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @return 1 if the wall is present, 0 if not present, or -1 if unknown
     */
    public int getVerticalState(int row, int column) {
        if (row < 0 || column < 0 || row >= height || column > width)
            throw new IllegalArgumentException("Invalid location for vertical wall: " + row + ", " + column + ".");
        return vertical[row][column];
    }
    
    /**
     * Adds or removes a horizontal wall.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @param state  1 to add the wall, 0 to remove, or -1 to set unknown
     */
    public void setHorizontalState(int row, int column, int state) {
        if (row < 0 || column < 0 || row > height || column >= width)
            throw new IllegalArgumentException("Invalid location for horizontal wall: " + row + ", " + column + ".");
        if (state < -1 || state > 1)
            throw new IllegalArgumentException("Invalid state: " + state + ".");
        horizontal[row][column] = state;
    }
    
    /**
     * Adds or removes a vertical wall.
     * 
     * @param row    the row index (from 0)
     * @param column the column index (from 0)
     * @param state  1 to add the wall, 0 to remove, or -1 to set unknown
     */
    public void setVerticalState(int row, int column, int state) {
        if (row < 0 || column < 0 || row >= height || column > width)
            throw new IllegalArgumentException("Invalid location for vertical wall: " + row + ", " + column + ".");
        if (state < -1 || state > 1)
            throw new IllegalArgumentException("Invalid state: " + state + ".");
        vertical[row][column] = state;
    }
    
    /**
     * Clears the state of the puzzle.
     */
    public void clearState() {
        for (int i=0; i<height+1; i++) for (int j=0; j<width; j++)
            horizontal[i][j] = -1;
        for (int i=0; i<height; i++) for (int j=0; j<width+1; j++)
            vertical[i][j] = -1;
    }
    
    //TODO
    public boolean isSolved() {
        return false;
    }
    
    /**
     * Generates a {@code String} representing the current situation of the
     * puzzle. Represents a wall as '|' or '-', corners as '+', no wall as ' ',
     * and unsure as '?'.
     * 
     * @return a {@code String} representing the state
     */
    public String toString() {
        int l = 2 * width + 2;
        char[] result = new char[l * (2 * height + 1)];
        for (int i=0; i<=height; i++) {
            for (int j=0; j<=width; j++) {
                result[i * 2 * l + j * 2] = '+';
                if (j < width)
                    result[i * 2 * l + j * 2 + 1] = horizontal[i][j] == 1 ? '-' : horizontal[i][j] == 0 ? ' ' : '?';
            }
            result[i * 2 * l + width * 2 + 1] = '\n';
            if (i < height) {
                for (int j=0; j<=width; j++) {
                    result[(i * 2 + 1) * l + j * 2] = vertical[i][j] == 1 ? '|' : vertical[i][j] == 0 ? ' ' : '?';
                    if (j < width)
                        result[(i * 2 + 1) * l + j * 2 + 1] = puzzle.clues[i][j] == -1 ? ' ' : (char) (puzzle.clues[i][j] + '0');
                }
                result[(i * 2 + 1) * l + width * 2 + 1] = '\n';
            }
        }
        return new String(result);
    }
    
}
